@@ -1,5 +1,5 @@
 from __future__ import annotations
-from io import BytesIO
+from io import BufferedIOBase
 from typing import (
     Any,
     NewType,
@@ -10,34 +10,42 @@ from typing import (
     Callable,
     Literal,
     Annotated,
+    ClassVar,
 )
 import typing
 from typing_extensions import dataclass_transform
 import inspect
 import dataclasses
 from enum import IntEnum
-from struct import Struct as _Struct, error as StructError
+from struct import Struct as _Struct, error as _StructError
 from decimal import Decimal
 
 
-__version__ = "0.1.0"
+__version__ = "0.4.0"
+
+
+class BstructError(Exception):
+    pass
 
 
 T = TypeVar("T")
 
-BytesDecoder = Callable[[bytes], T]
-BytesEncoder = Callable[[T], bytes]
-
-Decoder = Callable[[Iterator[Any]], T]
-Encoder = Callable[[T, list[Any]], None]
-
 ByteOrder = Literal["big", "little"]
+BytesDecoder = Callable[[bytes, ByteOrder], T]
+BytesEncoder = Callable[[T, ByteOrder], bytes]
+
+Decoder = Callable[[Iterator[Any], ByteOrder], T]
+Encoder = Callable[[T, list[Any], ByteOrder], None]
 
 
-def _raw_insert(value: Any, attributes: list[Any]) -> None:
+def _raw_encode(value: Any, attributes: list[Any], _byte_order: ByteOrder) -> None:
     attributes.append(value)
 
 
+def _raw_decode(iterator: Iterator[Any], _byteorder: ByteOrder) -> Any:
+    return next(iterator)
+
+
 class _StructEncoding(Generic[T]):
     def __init__(
         self,
@@ -46,8 +54,9 @@ class _StructEncoding(Generic[T]):
         encode: Encoder[T],
     ):
         self.format = format
-        self.struct = _Struct(f"<{format}")
-        self.size = self.struct.size
+        self.le_struct = _Struct(f"<{format}")
+        self.be_struct = _Struct(f">{format}")
+        self.size = self.le_struct.size
 
         self.decode = decode
         self.encode = encode
@@ -58,12 +67,18 @@ class _StructEncoding(Generic[T]):
     ) -> _StructEncoding[T]:
         return _StructEncoding(
             format=f"{size}s",
-            decode=lambda iterator: decode(next(iterator)),
-            encode=lambda t, l: l.append(encode(t)),
+            decode=lambda iterator, byte_order: decode(next(iterator), byte_order),
+            encode=lambda t, l, b: l.append(encode(t, b)),
         )
 
+    def get_struct(self, byteorder: ByteOrder) -> _Struct:
+        if byteorder == "little":
+            return self.le_struct
+        else:
+            return self.be_struct
 
-class CustomEncoding(Generic[T]):
+
+class _CustomEncoding(Generic[T]):
     def __init__(
         self,
         format: str,
@@ -76,56 +91,59 @@ class CustomEncoding(Generic[T]):
 
 
 class _NativeEncoding(Generic[T]):
+    decode: Decoder[T]
+    encode: Encoder[T]
+
     def __init__(self, format: str):
         self.format = format
-        self.decode = next
-        self.encode = _raw_insert
+        self.decode = _raw_decode
+        self.encode = _raw_encode
 
 
 class _IntEncoding(_NativeEncoding[int]):
     pass
 
 
-Encoding = Union[_NativeEncoding[T], CustomEncoding[T], _StructEncoding[T]]
+Encoding = Union[_NativeEncoding[T], _CustomEncoding[T], _StructEncoding[T]]
 
 
-def _decode_str(attributes: Iterator[Any]) -> str:
+def _decode_str(attributes: Iterator[Any], _byteorder: ByteOrder) -> str:
     value: bytes = next(attributes)
     return value.rstrip(b"\0").decode("utf-8")
 
 
-def _encode_str(value: str, attributes: list[Any]) -> None:
+def _encode_str(value: str, attributes: list[Any], _byteorder: ByteOrder) -> None:
     # The `struct` library automatically adds zeros to the end of
     # the encoded string to fill the necessary `data.size` bytes.
     data = value.encode("utf-8")
     attributes.append(data)
 
 
-def _decode_uint(attributes: Iterator[Any]) -> int:
-    return int.from_bytes(next(attributes), "little", signed=False)
+def _decode_uint(attributes: Iterator[Any], byteorder: ByteOrder) -> int:
+    return int.from_bytes(next(attributes), byteorder, signed=False)
 
 
-def _encode_uint128(value: int, attributes: list[Any]) -> None:
-    data = value.to_bytes(16, byteorder="little", signed=False)
+def _encode_uint128(value: int, attributes: list[Any], byteorder: ByteOrder) -> None:
+    data = value.to_bytes(16, byteorder, signed=False)
     attributes.append(data)
 
 
-def _encode_uint256(value: int, attributes: list[Any]) -> None:
-    data = value.to_bytes(32, byteorder="little", signed=False)
+def _encode_uint256(value: int, attributes: list[Any], byteorder: ByteOrder) -> None:
+    data = value.to_bytes(32, byteorder, signed=False)
     attributes.append(data)
 
 
-def _decode_int(attributes: Iterator[Any]) -> int:
-    return int.from_bytes(next(attributes), "little", signed=True)
+def _decode_int(attributes: Iterator[Any], byteorder: ByteOrder) -> int:
+    return int.from_bytes(next(attributes), byteorder, signed=True)
 
 
-def _encode_int128(value: int, attributes: list[Any]) -> None:
-    data = value.to_bytes(16, byteorder="little", signed=True)
+def _encode_int128(value: int, attributes: list[Any], byteorder: ByteOrder) -> None:
+    data = value.to_bytes(16, byteorder, signed=True)
     attributes.append(data)
 
 
-def _encode_int256(value: int, attributes: list[Any]) -> None:
-    data = value.to_bytes(32, byteorder="little", signed=True)
+def _encode_int256(value: int, attributes: list[Any], byteorder: ByteOrder) -> None:
+    data = value.to_bytes(32, byteorder, signed=True)
     attributes.append(data)
 
 
@@ -134,12 +152,14 @@ def _encode_int256(value: int, attributes: list[Any]) -> None:
 I80F48_DIVISOR = Decimal(2**48)
 
 
-def _decode_I80F48(attributes: Iterator[Any]) -> Decimal:
-    return int.from_bytes(next(attributes), "little", signed=True) / I80F48_DIVISOR
+def _decode_I80F48(attributes: Iterator[Any], byte_order: ByteOrder) -> Decimal:
+    return int.from_bytes(next(attributes), byte_order, signed=True) / I80F48_DIVISOR
 
 
-def _encode_I80F48(value: Decimal, attributes: list[Any]) -> None:
-    data = int(value * I80F48_DIVISOR).to_bytes(16, "little", signed=True)
+def _encode_I80F48(
+    value: Decimal, attributes: list[Any], byte_order: ByteOrder
+) -> None:
+    data = int(value * I80F48_DIVISOR).to_bytes(16, byte_order, signed=True)
     attributes.append(data)
 
 
@@ -150,10 +170,10 @@ class Encodings:
     u16 = _IntEncoding(format="H")
     u32 = _IntEncoding(format="I")
     u64 = _IntEncoding(format="Q")
-    u128: CustomEncoding[int] = CustomEncoding(
+    u128: _CustomEncoding[int] = _CustomEncoding(
         format="16s", decode=_decode_uint, encode=_encode_uint128
     )
-    u256: CustomEncoding[int] = CustomEncoding(
+    u256: _CustomEncoding[int] = _CustomEncoding(
         format="32s", decode=_decode_uint, encode=_encode_uint256
     )
 
@@ -161,14 +181,14 @@ class Encodings:
     i16 = _IntEncoding(format="h")
     i32 = _IntEncoding(format="i")
     i64 = _IntEncoding(format="q")
-    i128: CustomEncoding[int] = CustomEncoding(
+    i128: _CustomEncoding[int] = _CustomEncoding(
         format="16s", decode=_decode_int, encode=_encode_int128
     )
-    i256: CustomEncoding[int] = CustomEncoding(
+    i256: _CustomEncoding[int] = _CustomEncoding(
         format="32s", decode=_decode_int, encode=_encode_int256
     )
 
-    I80F48: CustomEncoding[Decimal] = CustomEncoding(
+    I80F48: _CustomEncoding[Decimal] = _CustomEncoding(
         "16s", _decode_I80F48, _encode_I80F48
     )
 
@@ -244,13 +264,15 @@ def _resolve_encoding(
     assert False
 
 
-def _encode_native_list(l: list[Any], attributes: list[Any]) -> None:
+def _encode_native_list(
+    l: list[Any], attributes: list[Any], _byteorder: ByteOrder
+) -> None:
     attributes.extend(l)
 
 
 def _resolve_array_encoding(
     array_type: Any, metadata: list[Any]
-) -> CustomEncoding[list[Any]]:
+) -> _CustomEncoding[list[Any]]:
     array_length = _resolve_array_length(metadata)
 
     array_type_args = typing.get_args(array_type)
@@ -263,20 +285,20 @@ def _resolve_array_encoding(
     inner_decode = inner_encoding.decode
     inner_encode = inner_encoding.encode
 
-    def _decode(attributes: Iterator[Any]) -> list[Any]:
-        return [inner_decode(attributes) for _ in range(array_length)]
+    def _decode(attributes: Iterator[Any], byte_order: ByteOrder) -> list[Any]:
+        return [inner_decode(attributes, byte_order) for _ in range(array_length)]
 
     # Optimization: If the encoding function of the element is just inserting the raw element,
     # use a batch insert to speed things up.
-    if inner_encode == _raw_insert:
+    if inner_encode == _raw_encode:
         _encode = _encode_native_list
     else:
 
-        def _encode(l: list[Any], attributes: list[Any]) -> None:
+        def _encode(l: list[Any], attributes: list[Any], byte_order: ByteOrder) -> None:
             for item in l:
-                inner_encode(item, attributes)
+                inner_encode(item, attributes, byte_order)
 
-    return CustomEncoding(array_format, _decode, _encode)
+    return _CustomEncoding(array_format, _decode, _encode)
 
 
 def _resolve_array_length(metadata: list[Any]) -> int:
@@ -289,29 +311,29 @@ def _resolve_array_length(metadata: list[Any]) -> int:
 
 def _resolve_int_enum_encoding(
     cls: Callable[[int], IntEnum], metadata: list[Any]
-) -> CustomEncoding[IntEnum]:
+) -> _CustomEncoding[IntEnum]:
     inner_encoding: Encoding[int] = _resolve_simple_encoding(metadata)
 
-    return CustomEncoding(
+    return _CustomEncoding(
         format=inner_encoding.format,
-        decode=lambda a: cls(inner_encoding.decode(a)),
+        decode=lambda a, b: cls(inner_encoding.decode(a, b)),
         encode=inner_encoding.encode,
     )
 
 
 def _resolve_simple_encoding(metadata: list[Any]) -> Encoding[Any]:
     for data in metadata:
-        if isinstance(data, (_IntEncoding, CustomEncoding)):
+        if isinstance(data, (_IntEncoding, _CustomEncoding)):
             data: Encoding[Any] = data
             return data
 
     raise TypeError("Cannot find integer type annotation")
 
 
-def _resolve_str_encoding(metadata: list[Any]) -> CustomEncoding[str]:
+def _resolve_str_encoding(metadata: list[Any]) -> _CustomEncoding[str]:
     for data in metadata:
         if isinstance(data, Size):
-            return CustomEncoding(
+            return _CustomEncoding(
                 format=f"{data.size}s",
                 decode=_decode_str,
                 encode=_encode_str,
@@ -330,16 +352,24 @@ def _resolve_bytes_encoding(metadata: list[Any]) -> _NativeEncoding[bytes]:
 
 @dataclass_transform()
 class Struct:
+    """
+    Inherit from this class to automatically derive the necessary
+    decoding/encoding information.
+    Also transforms any subclasses into a Python dataclass.
+    """
+
+    __bstruct_encoding__: ClassVar[_StructEncoding[Any]]
+
     def __init_subclass__(cls, **kwargs: Any) -> None:
         super().__init_subclass__(**kwargs)
 
         assert not dataclasses.is_dataclass(cls)
         dataclasses.dataclass(cls)
 
-        _derive(cls, byte_order="little")
+        cls.__bstruct_encoding__ = _derive(cls)
 
 
-def _derive(cls: type[T], byte_order: ByteOrder) -> _StructEncoding[T]:
+def _derive(cls: type[T]) -> _StructEncoding[T]:
     fields = dataclasses.fields(cls)
 
     attribute_decoders: list[Decoder[Any]] = []
@@ -354,35 +384,41 @@ def _derive(cls: type[T], byte_order: ByteOrder) -> _StructEncoding[T]:
 
     attribute_names = [field.name for field in fields]
 
-    def _decode(raw_attributes: Iterator[Any]) -> T:
+    def _decode(raw_attributes: Iterator[Any], byteorder: ByteOrder) -> T:
         attributes = [
-            decode_attribute(raw_attributes) for decode_attribute in attribute_decoders
+            decode_attribute(raw_attributes, byteorder)
+            for decode_attribute in attribute_decoders
         ]
 
         return cls(*attributes)
 
-    def _encode(value: T, raw_attributes: list[Any]) -> None:
+    def _encode(value: T, raw_attributes: list[Any], byteorder: ByteOrder) -> None:
         for encode_attribute, name in zip(attribute_encoders, attribute_names):
             attribute = getattr(value, name)
-            encode_attribute(attribute, raw_attributes)
+            encode_attribute(attribute, raw_attributes, byteorder)
 
-    encoding = _StructEncoding(
+    return _StructEncoding(
         format=full_format,
         decode=_decode,
         encode=_encode,
     )
-    _set_struct_encoding(cls, encoding)
-
-    return encoding
 
 
 def compile_format(fields: list[_NativeEncoding[Any]]) -> str:
+    """
+    Compile a list of attribute descriptions into a `struct.Struct` format string.
+    This does not contain the byteorder specifier (e.g. `<` or `>`).
+    """
     return "".join([field.format for field in fields])
 
 
 def patch(
     cls: type[T], size: int, decode: BytesDecoder[T], encode: BytesEncoder[T]
 ) -> None:
+    """
+    Patch an existing class.
+    The size must be the expected size in bytes of the serialized form of `T`.
+    """
     encoding = _StructEncoding.create_patched_data(size, decode, encode)
     _set_struct_encoding(cls, encoding)
 
@@ -390,72 +426,92 @@ def patch(
 S = TypeVar("S", bound=Struct)
 
 
-def decode(cls: type[S], data: bytes, strict: bool = True) -> S:
-    encoding = _get_struct_encoding(cls)
-
-    if not strict:
-        data = data[: encoding.size]
+def decode(cls: type[S], data: bytes, byteorder: ByteOrder = "little") -> S:
+    """
+    Decode an instance of `S` from the provided data.
+    The size of `data` must exactly match the size of `S`.
+    """
+    encoding = cls.__bstruct_encoding__
 
     try:
-        raw_attributes = encoding.struct.unpack(data)
-        return encoding.decode(iter(raw_attributes))
-    except StructError:
-        # TODO: Handle
-        raise
+        raw_attributes = encoding.get_struct(byteorder).unpack(data)
+        return encoding.decode(iter(raw_attributes), byteorder)
+    except _StructError as error:
+        raise BstructError(str(error)) from error
 
 
-def decode_all(cls: type[S], data: bytes) -> Iterator[S]:
-    encoding = _get_struct_encoding(cls)
+def decode_all(
+    cls: type[S], data: bytes, byteorder: ByteOrder = "little"
+) -> Iterator[S]:
+    """
+    Decode multiple instances of `S`.
+    The size of `data` must be an integer multiple of the size of `S`.
+    """
+    encoding = cls.__bstruct_encoding__
 
     try:
-        iterator = encoding.struct.iter_unpack(data)
+        iterator = encoding.get_struct(byteorder).iter_unpack(data)
 
         for raw_attributes in iterator:
-            yield encoding.decode(iter(raw_attributes))
-    except StructError:
-        # TODO: Handle
-        raise
+            yield encoding.decode(iter(raw_attributes), byteorder)
+    except _StructError as error:
+        raise BstructError(str(error)) from error
 
 
-def decode_from(cls: type[S], data_stream: BytesIO) -> S:
-    encoding = _get_struct_encoding(cls)
+def decode_from(
+    cls: type[S], data_stream: BufferedIOBase, byteorder: ByteOrder = "little"
+) -> S:
+    """
+    Read and decode an instance of `S` from the data stream.
+    """
+    encoding = cls.__bstruct_encoding__
 
     data = data_stream.read(encoding.size)
 
     try:
-        raw_attributes = encoding.struct.unpack(data)
-        return encoding.decode(iter(raw_attributes))
-    except StructError:
-        # TODO: Handle
-        raise
+        raw_attributes = encoding.get_struct(byteorder).unpack(data)
+        return encoding.decode(iter(raw_attributes), byteorder)
+    except _StructError as error:
+        raise BstructError(str(error)) from error
 
 
-def encode(value: Struct) -> bytes:
-    encoding = _get_struct_encoding(type(value))
+def encode(value: Struct, byteorder: ByteOrder = "little") -> bytes:
+    """
+    Encode the value according to the provided byteorder.
+    """
+    encoding = value.__bstruct_encoding__
 
     raw_attributes: list[Any] = []
-    encoding.encode(value, raw_attributes)
+    encoding.encode(value, raw_attributes, byteorder)
 
-    return encoding.struct.pack(*raw_attributes)
+    try:
+        return encoding.get_struct(byteorder).pack(*raw_attributes)
+    except _StructError as error:
+        raise BstructError(str(error)) from error
 
 
-def get_struct(cls: type[Struct]) -> _Struct:
-    encoding = _get_struct_encoding(cls)
-    return encoding.struct
+def get_struct(cls: type[Struct], byteorder: ByteOrder = "little") -> _Struct:
+    """
+    Return the underlying `struct.Struct` instance used to (un)pack
+    the binary data.
+    """
+    return cls.__bstruct_encoding__.get_struct(byteorder)
 
 
 def get_size(cls: type[Struct]) -> int:
-    encoding = _get_struct_encoding(cls)
-    return encoding.size
+    """
+    Return the size in bytes of the serialized form of a `bstruct.Struct` class.
+    """
+    return cls.__bstruct_encoding__.size
 
 
 def _get_struct_encoding(cls: type[S]) -> _StructEncoding[S]:
-    return getattr(cls, "__bstruct_encoding")
+    return getattr(cls, "__bstruct_encoding__")
 
 
 def _set_struct_encoding(cls: type[T], data: _StructEncoding[T]) -> None:
-    setattr(cls, "__bstruct_encoding", data)
+    setattr(cls, "__bstruct_encoding__", data)
 
 
 def _has_struct_encoding(cls: Any) -> bool:
-    return hasattr(cls, "__bstruct_encoding")
+    return hasattr(cls, "__bstruct_encoding__")
